"""Batch processing for tomography data with Warp and aretomo."""
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("waretomo")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Lorenzo Gaifas"
__email__ = "brisvag@gmail.com"



import os
import subprocess
from concurrent import futures

from rich import print


def run_threaded(
    progress,
    partials,
    label="",
    max_workers=None,
    dry_run=False,
    verbose=False,
    **kwargs,
):
    max_workers = max_workers or min(32, os.cpu_count() + 4)  # see concurrent docs

    with futures.ThreadPoolExecutor(max_workers) as executor:
        main_task = progress.add_task(f"{label}...", total=len(partials))

        jobs = []
        for fn in partials:
            job = executor.submit(fn)
            jobs.append(job)

        tasks = []
        for i, _thread in enumerate(executor._threads):
            tasks.append(progress.add_task(f"thread #{i}...", start=False))

        exist = 0
        errors = []
        for job in futures.as_completed(jobs):
            try:
                job.result()
            except FileExistsError:
                exist += 1
            except subprocess.CalledProcessError as e:
                errors.append(e)
                if verbose:
                    print(e)
            progress.update(main_task, advance=1)

        for t in tasks:
            progress.update(t, total=1, completed=1, visible=False)

        if exist:
            print(f"[red]{label}: {exist} files already exist and were not overwritten")

        if errors:
            print(f"[red]{label}: {len(errors)} commands have failed:")
            for err in errors:
                print(
                    f'[yellow]{" ".join(err.cmd)}[/yellow] '
                    f"failed with:\n[red]{err.stderr.decode()}"
                )



import os
import shutil
import subprocess
from contextlib import contextmanager
from pathlib import Path
from queue import Queue
from time import sleep

import GPUtil
from rich import print

from ._threaded import run_threaded


@contextmanager
def _cd(dir):
    prev = os.getcwd()
    os.chdir(dir)
    yield
    os.chdir(prev)


def _aretomo(
    input_,
    rawtlt,
    aln,
    xf,
    output,
    full_ts_name,
    suffix="",
    cmd="AreTomo",
    tilt_axis=0,
    patches=None,
    roi_file=None,
    tilt_corr=True,
    thickness_align=1200,
    thickness_recon=0,
    binning=4,
    px_size=0,
    kv=0,
    dose=0,
    cs=0,
    defocus=0,
    reconstruct=False,
    gpu_queue=None,
    dry_run=False,
    verbose=False,
    overwrite=False,
):
    # cwd dance is necessary cause aretomo messes up paths otherwise
    # need to use os.path.relpath cause pathlib cannot handle non-subpath relative paths
    # https://stackoverflow.com/questions/38083555/using-pathlibs-relative-to-for-directories-on-the-same-level
    cwd = output.parent.absolute()
    input_ = Path(os.path.relpath(input_, cwd))
    rawtlt = Path(os.path.relpath(rawtlt, cwd))
    aln = Path(os.path.relpath(aln, cwd))
    xf = Path(os.path.relpath(xf, cwd))
    output = Path(os.path.relpath(output, cwd))
    if not reconstruct:
        output = output.with_stem(output.stem + "_aligned").with_suffix(".st")
    # LogFile is broken, so we do it ourselves
    log = output.with_suffix(".aretomolog")
    with _cd(cwd):
        if not overwrite and output.exists():
            raise FileExistsError(output)

    # only one job per gpu
    gpu = 0 if gpu_queue is None else gpu_queue.get()

    options = {
        "InMrc": input_,
        "OutMrc": output,
        # 'LogFile': input_.with_suffix('.log').relative_to(cwd),  # currently broken
        "OutBin": binning,
        "Gpu": gpu,
        "DarkTol": 0,
    }

    if reconstruct:
        options.update(
            {
                "AlnFile": aln,
                "VolZ": thickness_recon,
                "PixSize": px_size,
                "Kv": kv,
                "ImgDose": dose,
                "Cs": cs,
                "Defoc": defocus,
                "FlipVol": 1,
                "WBP": 1,
            }
        )
    else:
        options.update(
            {
                "AngFile": rawtlt,
                "AlignZ": thickness_align,
                "TiltCor": int(tilt_corr),
                "VolZ": 0,
                "TiltAxis": f"{tilt_axis} 1" if tilt_axis is not None else "0 1",
                "OutImod": 2,
            }
        )
        if roi_file is not None:
            options["RoiFile"] = roi_file
        if patches is not None:
            options["Patch"] = f"{patches} {patches}"

    # run aretomo with basic settings
    aretomo_cmd = f"{cmd} {' '.join(f'-{k} {v}' for k, v in options.items())}"

    if verbose:
        print(aretomo_cmd)
        if not reconstruct:
            print(f'mv {xf} {full_ts_name + ".xf"}')

    if not dry_run:
        with _cd(cwd):
            try:
                proc = subprocess.run(
                    aretomo_cmd.split(), capture_output=True, check=False, cwd=cwd
                )
            finally:
                log.write_bytes(proc.stdout + proc.stderr)
                if gpu_queue is not None:
                    gpu_queue.put(gpu)
            proc.check_returncode()
        if not reconstruct:
            # move xf file so warp can see it (needs full ts name + .xf)
            shutil.move(cwd / xf, cwd / (full_ts_name + ".xf"))
    else:
        sleep(0.1)
        if gpu_queue is not None:
            gpu_queue.put(gpu)


def aretomo_batch(
    progress, tilt_series, suffix="", label="", cmd="AreTomo", gpus=None, **kwargs
):
    if not shutil.which(cmd):
        raise FileNotFoundError(f"{cmd} is not available on the system")
    if gpus is None:
        gpus = [gpu.id for gpu in GPUtil.getGPUs()]
    if not gpus:
        raise RuntimeError("you need at least one GPU to run AreTomo")
    if kwargs.get("verbose"):
        print(f"[yellow]Running AreTomo in parallel on {len(gpus)} GPUs.")

    # use a queue to hold gpu ids to ensure we only run one job per gpu
    gpu_queue = Queue()
    for gpu in gpus:
        gpu_queue.put(gpu)

    partials = []
    for ts in tilt_series:
        partials.append(
            lambda ts=ts: _aretomo(
                input_=ts["stack" + suffix] if suffix else ts["fix"],
                rawtlt=ts["rawtlt"],
                aln=ts["aln"],
                xf=ts["xf"],
                roi_file=ts["roi"],
                output=ts["recon" + suffix],
                gpu_queue=gpu_queue,
                cmd=cmd,
                full_ts_name=ts["name"],
                **ts["aretomo_kwargs"],
                **kwargs,
            )
        )

    run_threaded(progress, partials, label=label, max_workers=len(gpus), **kwargs)



import contextlib
import io
import multiprocessing
import re
import time
from concurrent import futures

from rich import print
from topaz.commands.denoise3d import denoise, load_model, set_device, train_model
from topaz.torch import set_num_threads


def _run_and_update_progress(progress, task, func, *args, **kwargs):
    std = io.StringIO()

    def _run_with_redirection():
        with contextlib.redirect_stderr(std):
            func(*args, **kwargs)

    with futures.ThreadPoolExecutor(1) as executor:
        job = executor.submit(_run_with_redirection)

        last_read_pos = 0
        while not job.done():
            time.sleep(0.5)
            std.seek(last_read_pos)
            last = std.read()
            if match := re.search(r"(\d+\.\d+)%", last):
                last_read_pos = std.tell()
                progress.update(task, completed=float(match.group(1)), refresh=True)

        try:
            return job.result()
        except RuntimeError as e:
            if "CUDA out of memory." in e.args[0]:
                raise RuntimeError(
                    "Not enough GPU memory. "
                    "Try to lower --topaz-tile-size or --topaz-patch-size"
                ) from e
            raise


def topaz_batch(
    progress,
    tilt_series,
    outdir,
    even,
    odd,
    model_name="unet-3d-10a",
    train=False,
    tile_size=32,
    patch_size=32,
    gpus=None,
    dry_run=False,
    verbose=False,
    overwrite=False,
):
    inputs = [ts["recon"] for ts in tilt_series]

    if verbose:
        if train:
            print(f"training model: '{model_name}' with inputs '{even}' and '{odd}'")
        if len(inputs) > 2:
            print(f"denoising: [{inputs[0]} [...] {inputs[-1]}]")
        else:
            print(f"denoising: {inputs}")
        print(f"output: {outdir}")

    if not dry_run:
        set_num_threads(0)
        std = io.StringIO()
        if train:
            task = progress.add_task(description="Training...")
            model, _ = _run_and_update_progress(
                progress,
                task,
                train_model,
                even_path=even,
                odd_path=odd,
                save_prefix=str(outdir / "trained_models" / model_name),
                save_interval=10,
                device=-2
                if gpus is None
                else next(iter(gpus)),  # TODO: actually run with multiple gpus?
                tilesize=patch_size,
                base_kernel_width=11,
                num_workers=multiprocessing.cpu_count(),
            )
        else:
            with contextlib.redirect_stdout(std):
                model = load_model(model_name, base_kernel_width=11)
        with contextlib.redirect_stdout(std):
            model.eval()
            model, use_cuda, num_devices = set_device(model, -2)

        for path in progress.track(inputs, description="Denoising..."):
            subtask = progress.add_task(description=path.name)
            _run_and_update_progress(
                progress,
                subtask,
                denoise,
                model=model,
                path=path,
                outdir=str(outdir),
                batch_size=num_devices,
                patch_size=patch_size,
                padding=patch_size // 2,
                suffix="",
            )
            progress.update(subtask, visible=False)



import shutil
import subprocess
from time import sleep

from rich import print

from ._threaded import run_threaded


def _stack(
    images, output, cmd="newstack", dry_run=False, verbose=False, overwrite=False
):
    if not overwrite and output.exists():
        raise FileExistsError(output)
    stack_cmd = f'{cmd} {" ".join(str(img) for img in images)} {output}'

    if verbose:
        short_cmd = f"{cmd} {images[0]} [...] {images[-1]} {output}"
        print(short_cmd)

    if not dry_run:
        subprocess.run(stack_cmd.split(), capture_output=True, check=True)
    else:
        sleep(0.1)


def prepare_half_stacks(progress, tilt_series, half, cmd="newstack", **kwargs):
    if not shutil.which(cmd):
        raise FileNotFoundError(f"{cmd} is not available on the system")

    partials = [
        lambda ts=ts: _stack(ts[half], ts[f"stack_{half}"], cmd=cmd, **kwargs)
        for ts in tilt_series
    ]
    run_threaded(progress, partials, label=f"Stacking {half} halves", **kwargs)



from enum import Enum, auto

import click

import waretomo


class ProcessingStep(str, Enum):
    """Enum for step selection."""

    fix = auto()
    align = auto()
    tilt_mdocs = auto()
    reconstruct = auto()
    stack_halves = auto()
    reconstruct_halves = auto()
    denoise = auto()

    def __str__(self):
        """String."""
        return self.name


@click.command(
    name="waretomo",
    context_settings={"help_option_names": ["-h", "--help"], "show_default": True},
)
@click.argument(
    "warp_dir", type=click.Path(exists=True, dir_okay=True, resolve_path=True)
)
@click.option(
    "-m",
    "--mdoc-dir",
    type=click.Path(exists=True, dir_okay=True, resolve_path=True),
    help="directory containing mdoc files [default: same as WARP_DIR]",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(exists=True, dir_okay=True, resolve_path=True),
    help="output directory for all the processing [default: WARP_DIR/waretomo]",
)
@click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    help="only print some info, without running the commands.",
)
@click.option("-v", "--verbose", is_flag=True, help="echo all individual commands")
@click.option(
    "-j",
    "--just",
    type=str,
    multiple=True,
    help="reconstruct just this tomogram (can be passed multiple times)",
)
@click.option(
    "-e",
    "--exclude",
    type=str,
    multiple=True,
    help="exclude this tomogram from the run (can be passed multiple times)",
)
@click.option(
    "-t",
    "--sample-thickness",
    type=int,
    default=400,
    help="unbinned thickness of the SAMPLE (ice or lamella) used for alignment",
)
@click.option(
    "-z",
    "--z-thickness",
    type=int,
    default=1200,
    help="unbinned thickness of the RECONSTRUCTION.",
)
@click.option(
    "-b",
    "--binning",
    type=int,
    default=4,
    help="binning for aretomo reconstruction (relative to warp pre-processed binning)",
)
@click.option(
    "-a", "--tilt-axis", type=float, help="starting tilt axis for AreTomo, if any"
)
@click.option(
    "-p",
    "--patches",
    type=int,
    help="number of patches for local alignment in aretomo (NxN), if any",
)
@click.option(
    "-r",
    "--roi-dir",
    type=click.Path(exists=True, dir_okay=True, resolve_path=True),
    help="directory containing ROI files. "
    "Extension does not matter, but names should be same as TS.",
)
@click.option(
    "-f", "--overwrite", is_flag=True, help="overwrite any previous existing run"
)
@click.option(
    "--train", is_flag=True, default=False, help="whether to train a new denosing model"
)
@click.option(
    "--topaz-tile-size",
    type=int,
    default=64,
    help="tile size for training topaz model.",
)
@click.option(
    "--topaz-patch-size",
    type=int,
    default=64,
    help="patch size for denoising in topaz.",
)
@click.option(
    "--topaz-model",
    type=str,
    default="unet-3d-10a",
    help="topaz model for denoising. If --train was not given, this must to be the "
    "name of a pre-trained model ('unet-3d-10a', 'unet-3d-20a') or the path of a "
    "previously generated model. If instead --train was given, a new model will"
    "be generated with the given name. New models are saved inside "
    "'OUTPUT_DIR/trained_models/.",
)
@click.option(
    "--start-from",
    type=click.Choice(ProcessingStep.__members__),
    default="fix",
    help="use outputs from a previous run, starting processing at this step",
)
@click.option(
    "--stop-at",
    type=click.Choice(ProcessingStep.__members__),
    default="reconstruct",
    help="terminate processing after this step",
)
@click.option("--ccderaser", type=str, default="ccderaser", help="ccderaser executable")
@click.option("--aretomo", type=str, default="AreTomo", help="aretomo executable")
@click.option(
    "--gpus",
    type=str,
    help="Comma separated list of gpus to use for aretomo. Default to all.",
)
@click.option(
    "--tiltcorr/--no-tiltcorr", default=True, help="do not correct sample tilt"
)
@click.version_option(waretomo.__version__, "-V", "--version")
def cli(
    warp_dir,
    mdoc_dir,
    output_dir,
    dry_run,
    verbose,
    just,
    exclude,
    sample_thickness,
    z_thickness,
    binning,
    tilt_axis,
    patches,
    roi_dir,
    overwrite,
    train,
    topaz_tile_size,
    topaz_patch_size,
    topaz_model,
    start_from,
    stop_at,
    ccderaser,
    aretomo,
    gpus,
    tiltcorr,
):
    """
    Run aretomo in batch on data preprocessed in warp.

    Needs to be ran after imod stacks were generated.
    Requires ccderaser and AreTomo>=1.3.0.

    Assumes the default Warp directory structure with generated imod stacks.
    """
    import sys
    from datetime import datetime
    from inspect import cleandoc
    from pathlib import Path

    from rich import print
    from rich.panel import Panel
    from rich.progress import Progress

    from ._parse import parse_data

    if gpus is not None:
        gpus = [int(gpu) for gpu in gpus.split(",")]

    warp_dir = Path(warp_dir)
    if mdoc_dir is None:
        mdoc_dir = warp_dir
    mdoc_dir = Path(mdoc_dir)
    if output_dir is None:
        output_dir = warp_dir / "waretomo_processing"
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    if roi_dir is not None:
        roi_dir = Path(roi_dir)

    pretrained_models = ("unet-3d-10a", "unet-3d-20a")
    if train:
        if topaz_model in pretrained_models:
            raise click.UsageError(
                f"Model '{topaz_model}' already exists. Provide a new name."
            )
        elif "/" in topaz_model:
            raise click.UsageError(
                f"Model '{topaz_model}' seems to be a path. "
                "Only a new name should be given."
            )
    else:
        if not Path(topaz_model).exists() and topaz_model not in pretrained_models:
            raise click.UsageError(
                f"Model '{topaz_model}' does not exist. "
                "Provide a path to an existing model "
                f"or one of {pretrained_models}."
            )

    with Progress() as progress:
        tilt_series, tilt_series_excluded, tilt_series_unprocessed = parse_data(
            progress,
            warp_dir,
            mdoc_dir=mdoc_dir,
            output_dir=output_dir,
            roi_dir=roi_dir,
            just=just,
            exclude=exclude,
            train=train,
        )

        aretomo_kwargs = {
            "cmd": aretomo,
            "tilt_axis": tilt_axis,
            "patches": patches,
            "thickness_align": sample_thickness,
            "thickness_recon": z_thickness,
            "binning": binning,
            "gpus": gpus,
            "tilt_corr": tiltcorr,
        }

        meta_kwargs = {
            "overwrite": overwrite,
            "dry_run": dry_run,
            "verbose": verbose,
        }

        topaz_kwargs = {
            "train": train,
            "model_name": topaz_model,
            "gpus": gpus,
            "tile_size": topaz_tile_size,
            "patch_size": topaz_patch_size,
        }

        start_from = ProcessingStep[start_from]
        stop_at = ProcessingStep[stop_at]

        steps = {
            step: start_from <= val <= stop_at
            for step, val in ProcessingStep.__members__.items()
        }
        if not train:
            steps["stack_halves"] = False
            steps["reconstruct_halves"] = False

        nl = "\n"

        not_ready_log = "".join(
            f'{nl}{" " * 12}- {ts}' for ts in tilt_series_unprocessed
        )
        ready_log = "".join(f'{nl}{" " * 12}- {ts["name"]}' for ts in tilt_series)
        excluded = "".join(f'{nl}{" " * 12}- {ts}' for ts in tilt_series_excluded)
        steps_log = "".join(
            f'{nl}{" " * 12}- '
            f'[{"green" if v else "red"}]{k}[/{"green" if v else "red"}] '
            for k, v in steps.items()
        )
        opts_log = "".join(f'{nl}{" " * 12}- {k}: {v}' for k, v in meta_kwargs.items())
        aretomo_opts_log = "".join(
            f'{nl}{" " * 12}- {k}: {v}' for k, v in aretomo_kwargs.items()
        )
        topaz_log = "".join(
            f'{nl}{" " * 12}- {k}: {v}' for k, v in topaz_kwargs.items()
        )

        summary = cleandoc(
            f"""
            [bold]Warp directory[/bold]: {warp_dir}
            [bold]Mdoc directory[/bold]: {mdoc_dir}
            [bold]Tilt series - NOT READY[/bold]: {not_ready_log}
            [bold]Tilt series - READY[/bold]: {ready_log}
            [bold]Tilt series - EXCLUDED[/bold]: {excluded}
            [bold]Processing steps[/bold]: {steps_log}
            [bold]Run options[/bold]: {opts_log}
            [bold]AreTomo options[/bold]: {aretomo_opts_log}
            [bold]Topaz options[/bold]: {topaz_log}
            """
        )
        print(Panel(summary))

        if not dry_run:
            with open(output_dir / "waretomo.log", "a") as f:
                print("=" * 80, file=f)
                print(datetime.now().strftime("%d/%m/%Y %H:%M:%S"), file=f)
                print(f'Command: {" ".join(sys.argv)}', file=f)
                print(summary, "\n", file=f)

        if steps["fix"]:
            from ._fix import fix_batch

            if verbose:
                print("\n[green]Fixing with ccderaser...")
            fix_batch(progress, tilt_series, cmd=ccderaser, **meta_kwargs)

        if steps["align"]:
            from ._aretomo import aretomo_batch

            if verbose:
                print("\n[green]Aligning with AreTomo...")
            aretomo_batch(
                progress,
                tilt_series,
                label="Aligning",
                **aretomo_kwargs,
                **meta_kwargs,
            )

        if steps["tilt_mdocs"]:
            if not tiltcorr:
                print("No need to tilt mdocs!")
            else:
                from ._fix_mdoc import tilt_mdocs_batch

                if verbose:
                    print("\n[green]Tilting mdocs...")
                (mdoc_dir / "mdoc_tilted").mkdir(parents=True, exist_ok=True)
                tilt_mdocs_batch(
                    progress,
                    tilt_series,
                    **meta_kwargs,
                )

        if steps["reconstruct"]:
            from ._aretomo import aretomo_batch

            if verbose:
                print("\n[green]Reconstructing with AreTomo...")
            aretomo_batch(
                progress,
                tilt_series,
                reconstruct=True,
                label="Reconstructing",
                **aretomo_kwargs,
                **meta_kwargs,
            )

        if steps["stack_halves"]:
            from ._stack import prepare_half_stacks

            for half in ("even", "odd"):
                if verbose:
                    print(f"\n[green]Preparing {half} stacks for denoising...")
                prepare_half_stacks(progress, tilt_series, half=half, **meta_kwargs)

        if steps["reconstruct_halves"]:
            from ._aretomo import aretomo_batch

            for half in ("even", "odd"):
                if verbose:
                    print(f"\n[green]Reconstructing {half} tomograms for deonoising...")
                half_dir = output_dir / half
                half_dir.mkdir(parents=True, exist_ok=True)
                aretomo_batch(
                    progress,
                    tilt_series,
                    suffix=f"_{half}",
                    reconstruct=True,
                    label=f"Reconstructing {half} halves",
                    **aretomo_kwargs,
                    **meta_kwargs,
                )
                # remove leftovers from aretomo otherwise topaz dies later
                for f in half_dir.glob("*_projX?.mrc"):
                    f.unlink(missing_ok=True)
                for f in half_dir.glob("*.aretomolog"):
                    f.unlink(missing_ok=True)

        if steps["denoise"]:
            from ._topaz import topaz_batch

            if verbose:
                print("\n[green]Denoising tomograms...")
            outdir_denoised = output_dir / "denoised"
            outdir_denoised.mkdir(parents=True, exist_ok=True)
            topaz_batch(
                progress,
                tilt_series,
                outdir=outdir_denoised,
                even=str(output_dir / "even"),
                odd=str(output_dir / "odd"),
                **topaz_kwargs,
                **meta_kwargs,
            )



from pathlib import Path, PureWindowsPath
from xml.etree import ElementTree

import mdocfile
from rich import print


def parse_data(
    progress, warp_dir, mdoc_dir, output_dir, roi_dir, just=(), exclude=(), train=False
):
    imod_dir = warp_dir / "imod"
    if not imod_dir.exists():
        raise FileNotFoundError("warp directory does not have an `imod` subdirectory")

    if just:
        mdocs = [
            p
            for ts_name in just
            if (p := (Path(mdoc_dir) / (ts_name + ".mdoc"))).exists()
        ]
    else:
        mdocs = sorted(Path(mdoc_dir).glob("*.mdoc"))

    if not mdocs:
        raise FileNotFoundError("could not find any mdoc files")

    odd_dir = warp_dir / "average" / "odd"
    even_dir = warp_dir / "average" / "even"

    tilt_series = []
    tilt_series_excluded = []
    tilt_series_unprocessed = []

    for mdoc in progress.track(mdocs, description="Reading mdocs..."):
        ts_name = mdoc.stem
        stack = imod_dir / ts_name / (ts_name + ".st")

        if ts_name in exclude:
            tilt_series_excluded.append(ts_name)
            continue
        # skip if not preprocessed in warp
        if not stack.exists():
            tilt_series_unprocessed.append(ts_name)
            continue

        df = mdocfile.read(mdoc)

        # extract even/odd paths
        tilts = [warp_dir / PureWindowsPath(tilt).name for tilt in df.SubFramePath]
        skipped_tilts = []
        odd = []
        even = []
        valid_xml = None
        for i, tilt in enumerate(tilts):
            if not tilt.exists():
                print(
                    f"[red]WARN: {tilt.name} is listed in an mdoc file, "
                    "but the file does not exists."
                )
                print(
                    "[red]      The tilt will be skipped, "
                    "but you may want to check your data."
                )
                skipped_tilts.append(i)
                continue

            xml = ElementTree.parse(tilt.with_suffix(".xml")).getroot()
            if xml.attrib["UnselectManual"] == "True":
                skipped_tilts.append(i)
            else:
                valid_xml = xml
                odd.append(odd_dir / (tilt.stem + ".mrc"))
                even.append(even_dir / (tilt.stem + ".mrc"))

        if valid_xml is None:
            tilt_series_unprocessed.append(ts_name)
            continue

        if train:
            for img in odd + even:
                if not img.exists():
                    raise FileNotFoundError(img)

        # extract metadata from warp xmls
        # (we assume the last xml has the same data as the others)
        for param in valid_xml.find("OptionsCTF"):
            if param.get("Name") == "BinTimes":
                binning = float(param.get("Value"))
            elif param.get("Name") == "Voltage":
                kv = int(param.get("Value"))
            elif param.get("Name") == "Cs":
                cs = float(param.get("Value"))
        for param in xml.find("CTF"):
            if param.get("Name") == "Defocus":
                defocus = (
                    float(param.get("Value")) * 1e4
                )  # defocus for aretomo is in Angstrom

        if roi_dir is not None:
            roi_files = list(roi_dir.glob(f"{ts_name}*"))
            if len(roi_files) == 1:
                roi_file = roi_files[0]
            else:
                roi_file = None
        else:
            roi_file = None

        ts_fixed = ts_name + "_fix"
        ts_stripped = ts_name.split(".")[0]
        alignment_result_dir = output_dir / (ts_stripped + "_Imod")

        tilt_series.append(
            {
                "name": ts_name,
                "stack": stack,
                "rawtlt": stack.with_suffix(".rawtlt"),
                "fix": output_dir / (ts_fixed + ".st"),
                "aln": output_dir / (ts_stripped + ".aln"),
                "xf": alignment_result_dir / (ts_stripped + ".xf"),
                "tlt": alignment_result_dir / (ts_stripped + ".tlt"),
                "skipped_tilts": skipped_tilts,
                "mdoc": mdoc,
                "roi": roi_file,
                "odd": odd,
                "even": even,
                "stack_odd": output_dir / (ts_name + "_odd.st"),
                "stack_even": output_dir / (ts_name + "_even.st"),
                "recon_odd": output_dir / "odd" / (ts_name + ".mrc"),
                "recon_even": output_dir / "even" / (ts_name + ".mrc"),
                "recon": output_dir / (ts_name + ".mrc"),
                "aretomo_kwargs": {
                    "dose": df.ExposureDose[0],
                    "px_size": df.PixelSpacing[0] * 2**binning,
                    "cs": cs,
                    "kv": kv,
                    "defocus": defocus,
                },
            }
        )

    return tilt_series, tilt_series_excluded, tilt_series_unprocessed



import pandas as pd
from mdocfile.data_models import Mdoc

from ._threaded import run_threaded


def _tilt_mdoc(
    mdoc_file, tlt_file, skipped_tilts, verbose=False, dry_run=False, overwrite=False
):
    output = mdoc_file.parent / "mdoc_tilted" / mdoc_file.name

    if not overwrite and output.exists():
        raise FileExistsError(output)

    if verbose:
        print(f"Tilting mdoc: {mdoc_file}")
        print(f"using: {tlt_file}")
        print(f"saving to {output}")

    if not dry_run:
        mdoc = Mdoc.from_file(mdoc_file)
        new_angles = iter(pd.read_csv(tlt_file, header=None)[0])
        # mdoc is still out of order, but aretomo puts the new_angles in order
        # this gives us tuples like (original_index, section), but ordered by angle
        # like aretomo so we can compare it with the new angles
        sorted_section_data = sorted(
            enumerate(mdoc.section_data), key=lambda x: x[1].TiltAngle
        )
        for idx, section in sorted_section_data:
            if idx in skipped_tilts:
                continue
            try:
                section.TiltAngle = next(new_angles)
            except StopIteration as e:
                raise RuntimeError(
                    f"not enough tilts generated by aretomo in {tlt_file}, "
                    f"or Mdoc is corrupted"
                ) from e

        try:
            next(new_angles)
        except StopIteration:
            pass
        else:
            raise RuntimeError(f"too many tilts generated by aretomo in {tlt_file}")

        with open(output, "w+" if overwrite else "w") as f:
            f.write(mdoc.to_string())


def tilt_mdocs_batch(progress, tilt_series, **kwargs):
    partials = []
    for ts in tilt_series:
        partials.append(
            lambda ts=ts: _tilt_mdoc(
                mdoc_file=ts["mdoc"],
                tlt_file=ts["tlt"],
                skipped_tilts=ts["skipped_tilts"],
                **kwargs,
            )
        )

    run_threaded(progress, partials, label="Creating tilted mdocs...", **kwargs)



import shutil
import subprocess
from time import sleep

from rich import print

from ._threaded import run_threaded


def _ccderaser(
    input, output, cmd="ccderaser", dry_run=False, verbose=False, overwrite=False
):
    if not overwrite and output.exists():
        raise FileExistsError(output)
    # run ccderaser, defaults from etomo
    ccderaser_cmd = (
        f"{cmd} -input {input} -output {output} -find -peak 8.0 "
        f"-diff 6.0 -big 19. -giant 12. -large 8. -grow 4. -edge 4"
    )

    if verbose:
        print(ccderaser_cmd)
    if not dry_run:
        subprocess.run(ccderaser_cmd.split(), capture_output=True, check=True)
    else:
        sleep(0.1)


def fix_batch(progress, tilt_series, cmd="ccderaser", **kwargs):
    if not shutil.which(cmd):
        raise FileNotFoundError(f"{cmd} is not available on the system")

    partials = [
        lambda ts=ts: _ccderaser(ts["stack"], ts["fix"], cmd=cmd, **kwargs)
        for ts in tilt_series
    ]
    run_threaded(progress, partials, label="Fixing", **kwargs)
